@@ -122,9 +122,9 @@ static Atom InitWcmAtom(DeviceIntPtr dev, char *name, Atom type, int format, int
 {
 	int i;
 	Atom atom;
-	uint8_t val_8[WCM_MAX_MOUSE_BUTTONS];
-	uint16_t val_16[WCM_MAX_MOUSE_BUTTONS];
-	uint32_t val_32[WCM_MAX_MOUSE_BUTTONS];
+	uint8_t val_8[WCM_MAX_BUTTONS];
+	uint16_t val_16[WCM_MAX_BUTTONS];
+	uint32_t val_32[WCM_MAX_BUTTONS];
 	pointer converted = val_32;
 
 	for (i = 0; i < nvalues; i++)
@@ -156,7 +156,8 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 {
 	WacomDevicePtr priv = (WacomDevicePtr) pInfo->private;
 	WacomCommonPtr common = priv->common;
-	int values[WCM_MAX_MOUSE_BUTTONS];
+	int values[WCM_MAX_BUTTONS];
+	int nbuttons;
 
 	DBG(10, priv, "\n");
 
@@ -189,7 +190,8 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	values[1] = priv->old_serial;
 	values[2] = priv->old_device_id;
 	values[3] = priv->cur_serial;
-	prop_serials = InitWcmAtom(pInfo->dev, WACOM_PROP_SERIALIDS, XA_INTEGER, 32, 4, values);
+	values[4] = priv->cur_device_id;
+	prop_serials = InitWcmAtom(pInfo->dev, WACOM_PROP_SERIALIDS, XA_INTEGER, 32, 5, values);
 
 	values[0] = priv->serial;
 	prop_serial_binding = InitWcmAtom(pInfo->dev, WACOM_PROP_SERIAL_BIND, XA_INTEGER, 32, 1, values);
@@ -225,9 +227,11 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	values[0] = MakeAtom(pInfo->type_name, strlen(pInfo->type_name), TRUE);
 	prop_tooltype = InitWcmAtom(pInfo->dev, WACOM_PROP_TOOL_TYPE, XA_ATOM, 32, 1, values);
 
+
 	/* default to no actions */
+	nbuttons = min(max(priv->nbuttons + 4, 7), WCM_MAX_BUTTONS);
 	memset(values, 0, sizeof(values));
-	prop_btnactions = InitWcmAtom(pInfo->dev, WACOM_PROP_BUTTON_ACTIONS, XA_ATOM, 32, WCM_MAX_MOUSE_BUTTONS, values);
+	prop_btnactions = InitWcmAtom(pInfo->dev, WACOM_PROP_BUTTON_ACTIONS, XA_ATOM, 32, nbuttons, values);
 
 	if (IsPad(priv)) {
 		memset(values, 0, sizeof(values));
@@ -237,7 +241,7 @@ void InitWcmDeviceProperties(InputInfoPtr pInfo)
 	if (IsPad(priv) || IsCursor(priv))
 	{
 		memset(values, 0, sizeof(values));
-		prop_wheel_buttons = InitWcmAtom(pInfo->dev, WACOM_PROP_WHEELBUTTONS, XA_ATOM, 32, 4, values);
+		prop_wheel_buttons = InitWcmAtom(pInfo->dev, WACOM_PROP_WHEELBUTTONS, XA_ATOM, 32, 6, values);
 	}
 
 	values[0] = common->vendor_id;
@@ -288,7 +292,7 @@ static int wcmSanityCheckProperty(XIPropertyValuePtr prop)
 			case AC_KEY:
 				break;
 			case AC_BUTTON:
-				if (code > WCM_MAX_MOUSE_BUTTONS)
+				if (code > WCM_MAX_BUTTONS)
 					return BadValue;
 				break;
 			case AC_DISPLAYTOGGLE:
@@ -456,6 +460,8 @@ struct wheel_strip_update_t {
 	int *dn1;
 	int *up2;
 	int *dn2;
+	int *up3;
+	int *dn3;
 
 	/* for CARD32 values, points to atom array of atoms to be
 	 * monitored.*/
@@ -477,7 +483,8 @@ static int wcmSetWheelOrStripProperty(DeviceIntPtr dev, Atom property,
 		CARD32 *v32;
 	} values;
 
-	if (prop->size != 4)
+	if ((property == prop_strip_buttons && prop->size != 4) ||
+	    (property == prop_wheel_buttons && prop->size != 6))
 		return BadValue;
 
 	/* see wcmSetPropertyButtonActions for how this works. The wheel is
@@ -489,10 +496,12 @@ static int wcmSetWheelOrStripProperty(DeviceIntPtr dev, Atom property,
 	switch (prop->format)
 	{
 		case 8:
-			if (values.v8[0] > WCM_MAX_MOUSE_BUTTONS ||
-			    values.v8[1] > WCM_MAX_MOUSE_BUTTONS ||
-			    values.v8[2] > WCM_MAX_MOUSE_BUTTONS ||
-			    values.v8[3] > WCM_MAX_MOUSE_BUTTONS)
+			if (values.v8[0] > WCM_MAX_BUTTONS ||
+			    values.v8[1] > WCM_MAX_BUTTONS ||
+			    values.v8[2] > WCM_MAX_BUTTONS ||
+			    values.v8[3] > WCM_MAX_BUTTONS ||
+			    values.v8[4] > WCM_MAX_BUTTONS ||
+			    values.v8[5] > WCM_MAX_BUTTONS)
 				return BadValue;
 
 			if (!checkonly) {
@@ -500,6 +509,8 @@ static int wcmSetWheelOrStripProperty(DeviceIntPtr dev, Atom property,
 				*wsup->dn1 = values.v8[1];
 				*wsup->up2 = values.v8[2];
 				*wsup->dn2 = values.v8[3];
+				*wsup->up3 = values.v8[4];
+				*wsup->dn3 = values.v8[5];
 			}
 			break;
 		case 32:
@@ -534,10 +545,12 @@ static int wcmSetWheelProperty(DeviceIntPtr dev, Atom property,
 		.dn1 = &priv->reldn,
 		.up2 = &priv->wheelup,
 		.dn2 = &priv->wheeldn,
+		.up3 = &priv->wheel2up,
+		.dn3 = &priv->wheel2dn,
 
 		.handlers = priv->wheel_actions,
 		.keys	  = priv->wheel_keys,
-		.skeys    = 4,
+		.skeys    = 6,
 	};
 
 	return wcmSetWheelOrStripProperty(dev, property, prop, checkonly, &wsup);
@@ -554,6 +567,8 @@ static int wcmSetStripProperty(DeviceIntPtr dev, Atom property,
 		.dn1 = &priv->stripldn,
 		.up2 = &priv->striprup,
 		.dn2 = &priv->striprdn,
+		.up3 = NULL,
+		.dn3 = NULL,
 
 		.handlers = priv->strip_actions,
 		.keys	  = priv->strip_keys,
@@ -704,7 +719,7 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 		 * set it at runtime. If we get here from wcmUpdateSerial,
 		 * we know the serial has ben set internally already, so we
 		 * can reply with success. */
-		if (prop->size == 4 && prop->format == 32)
+		if (prop->size == 5 && prop->format == 32)
 			if (((CARD32*)prop->data)[3] == priv->cur_serial)
 				return Success;
 
@@ -744,14 +759,16 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 			common->wcmCursorProxoutDist = value;
 	} else if (property == prop_threshold)
 	{
-		CARD32 value;
+		INT32 value;
 
 		if (prop->size != 1 || prop->format != 32)
 			return BadValue;
 
-		value = *(CARD32*)prop->data;
+		value = *(INT32*)prop->data;
 
-		if ((value < 1) || (value > FILTER_PRESSURE_RES))
+		if (value == -1)
+			value = DEFAULT_THRESHOLD;
+		else if ((value < 1) || (value > FILTER_PRESSURE_RES))
 			return BadValue;
 
 		if (!checkonly)
@@ -833,7 +850,8 @@ int wcmSetProperty(DeviceIntPtr dev, Atom property, XIPropertyValuePtr prop,
 #endif
 	} else if (property == prop_btnactions)
 	{
-		if (prop->size != WCM_MAX_MOUSE_BUTTONS)
+		int nbuttons = min(max(priv->nbuttons + 4, 7), WCM_MAX_BUTTONS);
+		if (prop->size != nbuttons)
 			return BadMatch;
 		wcmSetPropertyButtonActions(dev, property, prop, checkonly);
 	} else
@@ -852,17 +870,18 @@ int wcmGetProperty (DeviceIntPtr dev, Atom property)
 
 	if (property == prop_serials)
 	{
-		uint32_t values[4];
+		uint32_t values[5];
 
 		values[0] = common->tablet_id;
 		values[1] = priv->old_serial;
 		values[2] = priv->old_device_id;
 		values[3] = priv->cur_serial;
+		values[4] = priv->cur_device_id;
 
 		DBG(10, priv, "Update to serial: %d\n", priv->old_serial);
 
 		return XIChangeDeviceProperty(dev, property, XA_INTEGER, 32,
-					      PropModeReplace, 4,
+					      PropModeReplace, 5,
 					      values, FALSE);
 	}
 
@@ -875,14 +894,14 @@ serialTimerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
 	InputInfoPtr pInfo = arg;
 	WacomDevicePtr priv = pInfo->private;
 	XIPropertyValuePtr prop;
-	CARD32 prop_value[4];
+	CARD32 prop_value[5];
 	int sigstate;
 	int rc;
 
 	sigstate = xf86BlockSIGIO();
 
 	rc = XIGetDeviceProperty(pInfo->dev, prop_serials, &prop);
-	if (rc != Success || prop->format != 32 || prop->size != 4)
+	if (rc != Success || prop->format != 32 || prop->size != 5)
 	{
 		xf86Msg(X_ERROR, "%s: Failed to update serial number.\n",
 			pInfo->name);
@@ -891,6 +910,7 @@ serialTimerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
 
 	memcpy(prop_value, prop->data, sizeof(prop_value));
 	prop_value[3] = priv->cur_serial;
+	prop_value[4] = priv->cur_device_id;
 
 	XIChangeDeviceProperty(pInfo->dev, prop_serials, XA_INTEGER,
 			       prop->format, PropModeReplace,
@@ -902,7 +922,7 @@ serialTimerFunc(OsTimerPtr timer, CARD32 now, pointer arg)
 }
 
 void
-wcmUpdateSerial(InputInfoPtr pInfo, unsigned int serial)
+wcmUpdateSerial(InputInfoPtr pInfo, unsigned int serial, int id)
 {
 	WacomDevicePtr priv = pInfo->private;
 
@@ -910,6 +930,7 @@ wcmUpdateSerial(InputInfoPtr pInfo, unsigned int serial)
 		return;
 
 	priv->cur_serial = serial;
+	priv->cur_device_id = id;
 
 	/* This function is called during SIGIO. Schedule timer for property
 	 * event delivery outside of signal handler. */
