@@ -300,6 +300,26 @@ static param_t parameters[] =
 		.set_func = map_actions,
 		.get_func = get_map,
 	},
+	{
+		.name = "AbsWheel2Up",
+		.desc = "X11 event to which absolute wheel up should be mapped. ",
+		.prop_name = WACOM_PROP_WHEELBUTTONS,
+		.prop_format = 8,
+		.prop_offset = 4,
+		.arg_count = 0,
+		.set_func = map_actions,
+		.get_func = get_map,
+	},
+	{
+		.name = "AbsWheel2Down",
+		.desc = "X11 event to which absolute wheel down should be mapped. ",
+		.prop_name = WACOM_PROP_WHEELBUTTONS,
+		.prop_format = 8,
+		.prop_offset = 5,
+		.arg_count = 0,
+		.set_func = map_actions,
+		.get_func = get_map,
+	},
 	{
 		.name = "StripLeftUp",
 		.desc = "X11 event to which left strip up should be mapped. ",
@@ -360,8 +380,8 @@ static param_t parameters[] =
 		.set_func = set_xydefault,
 	},
 	{
-		.name = "ToolID",
-		.desc = "Returns the ID of the associated device. ",
+		.name = "ToolType",
+		.desc = "Returns the tool type of the associated device. ",
 		.prop_name = WACOM_PROP_TOOL_TYPE,
 		.prop_format = 32,
 		.prop_offset = 0,
@@ -377,6 +397,15 @@ static param_t parameters[] =
 		.arg_count = 1,
 		.prop_flags = PROP_FLAG_READONLY
 	},
+	{
+		.name = "ToolID",
+		.desc = "Returns the tool ID of the current tool in proximity.",
+		.prop_name = WACOM_PROP_SERIALIDS,
+		.prop_format = 32,
+		.prop_offset = 4,
+		.arg_count = 1,
+		.prop_flags = PROP_FLAG_READONLY
+	},
 	{
 		.name = "ToolSerialPrevious",
 		.desc = "Returns the serial number of the previous device in proximity.",
@@ -2327,7 +2356,6 @@ static void set_output(Display *dpy, XDevice *dev, param_t *param, int argc, cha
 	int x, y;
 	unsigned int width, height;
 	int flags = XParseGeometry(argv[0], &x, &y, &width, &height);
-	Bool success = False;
 
 	if (argc != param->arg_count)
 	{
@@ -2337,15 +2365,15 @@ static void set_output(Display *dpy, XDevice *dev, param_t *param, int argc, cha
 	}
 
 	if (MaskIsSet(flags, XValue|YValue|WidthValue|HeightValue))
-		success = set_output_area(dpy, dev, x, y, width, height);
+		set_output_area(dpy, dev, x, y, width, height);
 	else if (strcasecmp(argv[0], "next") == 0)
-		success = set_output_next(dpy, dev);
+		set_output_next(dpy, dev);
 	else if (strcasecmp(argv[0], "desktop") == 0)
-		success = set_output_desktop(dpy, dev);
+		set_output_desktop(dpy, dev);
 	else if (!need_xinerama(dpy))
-		success = set_output_xrandr(dpy, dev, argv[0]);
+		set_output_xrandr(dpy, dev, argv[0]);
 	else if  (convert_value_from_user(param, argv[0], &head_no))
-		success = set_output_xinerama(dpy, dev, head_no);
+		set_output_xinerama(dpy, dev, head_no);
 	else
 		fprintf(stderr, "Unable to find an output '%s'.\n", argv[0]);
 }
@@ -2669,7 +2697,7 @@ static void test_parameter_number(void)
 	 * deprecated them.
 	 * Numbers include trailing NULL entry.
 	 */
-	assert(ARRAY_SIZE(parameters) == 34);
+	assert(ARRAY_SIZE(parameters) == 37);
 	assert(ARRAY_SIZE(deprecated_parameters) == 17);
 }
 
